from datetime import datetime
from typing import Any, Dict, List, MutableMapping, Optional, TypeVar

KT = TypeVar("KT")
VT = TypeVar("VT")
Colors = Dict[str, int]
Colours = Colors


def camel_to_snake(string: str) -> str:
    return "".join(["_" + c.lower() if c.isupper() else c for c in string]).lstrip("_")


def parse_vote_dict(d: dict) -> dict:
    data = d.copy()

    query = data.get("query", "").lstrip("?")
    if query:
        query_dict = {k: v for k, v in [pair.split("=") for pair in query.split("&")]}
        data["query"] = DataDict(**query_dict)
    else:
        data["query"] = {}

    if "bot" in data:
        data["bot"] = int(data["bot"])

    elif "guild" in data:
        data["guild"] = int(data["guild"])

    for key, value in data.copy().items():
        converted_key = camel_to_snake(key)
        if key != converted_key:
            del data[key]
            data[converted_key] = value

    return data


def parse_dict(d: dict) -> dict:
    data = d.copy()

    for key, value in data.copy().items():
        if "id" in key.lower():
            if value == "":
                value = None
            else:
                if isinstance(value, str) and value.isdigit():
                    value = int(value)
                else:
                    continue
        elif value == "":
            value = None

        converted_key = camel_to_snake(key)
        if key != converted_key:
            del data[key]
        data[converted_key] = value

    return data


def parse_bot_dict(d: dict) -> dict:
    data = parse_dict(d.copy())

    if data.get("date") and not isinstance(data["date"], datetime):
        data["date"] = datetime.strptime(data["date"], "%Y-%m-%dT%H:%M:%S.%fZ")

    if data.get("owners"):
        data["owners"] = [int(e) for e in data["owners"]]
    if data.get("guilds"):
        data["guilds"] = [int(e) for e in data["guilds"]]

    for key, value in data.copy().items():
        converted_key = camel_to_snake(key)
        if key != converted_key:
            del data[key]
        data[converted_key] = value

    return data


def parse_user_dict(d: dict) -> dict:
    data = d.copy()

    data["social"] = SocialData(**data.get("social", {}))

    return data


def parse_bot_stats_dict(d: dict) -> dict:
    data = d.copy()

    if "server_count" not in data:
        data["server_count"] = None
    if "shards" not in data:
        data["shards"] = []
    if "shard_count" not in data:
        data["shard_count"] = None

    return data


class DataDict(dict, MutableMapping[KT, VT]):
    """Base class used to represent received data from the API.

    Every data model subclasses this class.
    """

    def __init__(self, **kwargs: VT) -> None:
        super().__init__(**parse_dict(kwargs))
        self.__dict__ = self


class WidgetOptions(DataDict[str, Any]):
    """Model that represents widget options that are passed to Top.gg widget URL generated via
    :meth:`DBLClient.generate_widget`."""

    id: Optional[int]
    """ID of a bot to generate the widget for. Must resolve to an ID of a listed bot when converted to a string."""
    colors: Colors
    """A dictionary consisting of a parameter as a key and HEX color (type `int`) as value. ``color`` will be 
    appended to the key in case it doesn't end with ``color``."""
    noavatar: bool
    """Indicates whether to exclude the bot's avatar from short widgets. Must be of type ``bool``. Defaults to 
    ``False``."""
    format: str
    """Format to apply to the widget. Must be either ``png`` and ``svg``. Defaults to ``png``."""
    type: str
    """Type of a short widget (``status``, ``servers``, ``library``, ``upvotes``, and ``owner``). For large widget, 
    must be an empty string."""

    def __init__(
        self,
        id: Optional[int] = None,
        format: Optional[str] = None,
        type: Optional[str] = None,
        noavatar: bool = False,
        colors: Optional[Colors] = None,
        colours: Optional[Colors] = None,
    ):
        super().__init__(
            id=id or None,
            format=format or "png",
            type=type or "",
            noavatar=noavatar or False,
            colors=colors or colours or {},
        )

    @property
    def colours(self) -> Colors:
        return self.colors

    @colours.setter
    def colours(self, value: Colors) -> None:
        self.colors = value

    def __setitem__(self, key: str, value: Any) -> None:
        if key == "colours":
            key = "colors"
        super().__setitem__(key, value)

    def __getitem__(self, item: str) -> Any:
        if item == "colours":
            item = "colors"
        return super().__getitem__(item)

    def get(self, key: str, default: Any = None) -> Any:
        """:meta private:"""
        if key == "colours":
            key = "colors"
        return super().get(key, default)


class BotData(DataDict[str, Any]):
    """Model that contains information about a listed bot on top.gg. The data this model contains can be found `here
    <https://docs.top.gg/api/bot/#bot-structure>`__."""

    id: int
    username: str
    discriminator: str
    avatar: Optional[str]
    def_avatar: str
    prefix: str
    shortdesc: str
    longdesc: Optional[str]
    tags: List[str]
    website: Optional[str]
    support: Optional[str]
    github: Optional[str]
    owners: List[int]
    guilds: List[int]
    invite: Optional[str]
    date: datetime
    certified_bot: bool
    vanity: Optional[str]
    points: int
    monthly_points: int
    donatebotguildid: int

    def __init__(self, **kwargs: Any):
        super().__init__(**parse_bot_dict(kwargs))


class BotStatsData(DataDict[str, Any]):
    """Model that contains information about a listed bot's guild and shard count."""

    server_count: Optional[int]
    """The amount of servers the bot is in."""
    shards: List[int]
    """The amount of servers the bot is in per shard."""
    shard_count: Optional[int]
    """The amount of shards a bot has."""

    def __init__(self, **kwargs: Any):
        super().__init__(**parse_bot_stats_dict(kwargs))


class BriefUserData(DataDict[str, Any]):
    """Model that contains brief information about a Top.gg user."""

    id: int
    """The Discord ID of the user."""
    username: str
    """The Discord username of the user."""
    avatar: str
    """The Discord avatar URL of the user."""

    def __init__(self, **kwargs: Any):
        if kwargs["id"].isdigit():
            kwargs["id"] = int(kwargs["id"])
        super().__init__(**kwargs)


class SocialData(DataDict[str, str]):
    """Model that contains social information about a top.gg user."""

    youtube: str
    """The YouTube channel ID of the user."""
    reddit: str
    """The Reddit username of the user."""
    twitter: str
    """The Twitter username of the user."""
    instagram: str
    """The Instagram username of the user."""
    github: str
    """The GitHub username of the user."""


class UserData(DataDict[str, Any]):
    """Model that contains information about a top.gg user. The data this model contains can be found `here
    <https://docs.top.gg/api/user/#structure>`__."""

    id: int
    username: str
    discriminator: str
    social: SocialData
    color: str
    supporter: bool
    certified_dev: bool
    mod: bool
    web_mod: bool
    admin: bool

    def __init__(self, **kwargs: Any):
        super().__init__(**parse_user_dict(kwargs))


class VoteDataDict(DataDict[str, Any]):
    """Base model that represents received information from Top.gg via webhooks."""

    type: str
    """Type of the action (``upvote`` or ``test``)."""
    user: int
    """ID of the voter."""
    query: DataDict
    """Query parameters in :ref:`DataDict`."""

    def __init__(self, **kwargs: Any):
        super().__init__(**parse_vote_dict(kwargs))


class BotVoteData(VoteDataDict):
    """Model that contains information about a bot vote."""

    bot: int
    """ID of the bot the user voted for."""
    is_weekend: bool
    """Boolean value indicating whether the action was done on a weekend."""


class ServerVoteData(VoteDataDict):
    """Model that contains information about a server vote."""

    guild: int
    """ID of the guild the user voted for."""



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2021 Assanali Mukhanov

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
import logging
import sys

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from typing import TYPE_CHECKING, Dict

import aiohttp
import discord
from aiohttp import web

from .types import BotVoteData, ServerVoteData

if TYPE_CHECKING:
    import asyncio

    # noinspection PyProtectedMember
    from aiohttp.web_urldispatcher import _WebHandler

log = logging.getLogger(__name__)


class _Webhook(TypedDict):
    route: str
    auth: str
    func: "_WebHandler"


class WebhookManager:
    """
    This class is used as a manager for the Top.gg webhook.

    Methods :meth:`WebhookManager.dbl_webhook` and :meth:`WebhookManager.dsl_webhook` return a modified version of
    the object, allowing for method chaining.

    Parameters
    ----------
    bot: discord.Client
        The Client object that will be utilized by this manager's webhook(s) to emit events.
    """

    __app: web.Application
    _webhooks: Dict[
        str,
        _Webhook,
    ]
    _webserver: web.TCPSite
    _is_closed: bool

    def __init__(self, bot: discord.Client):
        self.bot = bot
        self._webhooks = {}
        self.__app = web.Application()
        self._is_closed = False

    def dbl_webhook(self, route: str = "/dbl", auth_key: str = "") -> "WebhookManager":
        """Helper method that configures a route that listens to bot votes.

        Parameters
        ----------
        route: str
            The route to use for bot votes. Must start with ``/``. Defaults to ``/dbl``.
        auth_key: str
            The Authorization key that will be used to verify the incoming requests.
            All requests are allowed if this is not set.

        Returns
        -------
        Modified version of the object: WebhookManager
        """
        self._webhooks["dbl"] = _Webhook(
            route=route or "/dbl",
            auth=auth_key or "",
            func=self._bot_vote_handler,
        )
        return self

    def dsl_webhook(self, route: str = "/dsl", auth_key: str = "") -> "WebhookManager":
        """Helper method that configures a route that listens to server votes.

        Parameters
        ----------
        route: str
            The route to use for server votes. Must start with ``/``. Defaults to ``/dsl``.
        auth_key: str
            The Authorization key that will be used to verify the incoming requests.
            All requests are allowed if this is not set.

        Returns
        -------
        Modified version of the object: WebhookManager
        """
        self._webhooks["dsl"] = _Webhook(
            route=route or "/dsl",
            auth=auth_key or "",
            func=self._guild_vote_handler,
        )
        return self

    async def _bot_vote_handler(self, request: aiohttp.web.Request) -> web.Response:
        auth = request.headers.get("Authorization", "")
        if auth == self._webhooks["dbl"]["auth"]:
            data = await request.json()
            self.bot.dispatch("dbl_vote", BotVoteData(**data))
            return web.Response(status=200, text="OK")
        return web.Response(status=401, text="Unauthorized")

    async def _guild_vote_handler(self, request: aiohttp.web.Request) -> web.Response:
        auth = request.headers.get("Authorization", "")
        if auth == self._webhooks["dsl"]["auth"]:
            data = await request.json()
            self.bot.dispatch("dsl_vote", ServerVoteData(**data))
            return web.Response(status=200, text="OK")
        return web.Response(status=401, text="Unauthorized")

    async def _run(self, port: int) -> None:
        for webhook in self._webhooks.values():
            self.__app.router.add_post(webhook["route"], webhook["func"])
        runner = web.AppRunner(self.__app)
        await runner.setup()
        self._webserver = web.TCPSite(runner, "0.0.0.0", port)
        await self._webserver.start()
        self._is_closed = False

    def run(self, port: int) -> "asyncio.Task[None]":
        """Runs the webhook.

        Parameters
        ----------
        port: int
            The port to run the webhook on.
        """
        return self.bot.loop.create_task(self._run(port))

    @property
    def webserver(self) -> web.Application:
        """Returns the internal web application that handles webhook requests.

        Returns
        --------
        :class:`aiohttp.web.Application`
            The internal web application.
        """
        return self.__app

    async def close(self) -> None:
        """Stops the webhook."""
        await self._webserver.stop()
        self._is_closed = True



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2021 Assanali Mukhanov

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
import json
import logging
import sys
from datetime import datetime
from typing import Any, Coroutine, Dict, Iterable, List, Optional, Sequence, Union, cast

import aiohttp
from aiohttp import ClientResponse

from . import __version__, errors
from .ratelimiter import AsyncRateLimiter, AsyncRateLimiterManager

log = logging.getLogger(__name__)


async def _json_or_text(
    response: ClientResponse,
) -> Union[dict, str]:
    """

    Parameters
    ----------
    response: ClientResponse
        The received aiohttp response object.

    Returns
    -------
    body: Union[dict, str]
        Response body in either JSON or string.
    """
    text = await response.text()
    if response.headers["Content-Type"] == "application/json; charset=utf-8":
        return json.loads(text)
    return text


class HTTPClient:
    """Represents an HTTP client sending HTTP requests to the Top.gg API.

    .. _event loop: https://docs.python.org/3/library/asyncio-eventloops.html
    .. _aiohttp session: https://aiohttp.readthedocs.io/en/stable/client_reference.html#client-session

    Parameters
    ----------
    token:
        A Top.gg API Token.
    **session: `aiohttp session`_
        The `aiohttp session`_ used for requests to the API.
    **loop: `event loop`_
        An `event loop`_ used for asynchronous operations.
    """

    def __init__(self, token: str, **kwargs: Any) -> None:
        self.BASE = "https://top.gg/api"
        self.token = token
        self.loop = kwargs.get("loop") or asyncio.get_event_loop()
        self.session = kwargs.get("session") or aiohttp.ClientSession(loop=self.loop)
        self.global_rate_limiter = AsyncRateLimiter(
            max_calls=99, period=1, callback=_rate_limit_handler
        )
        self.bot_rate_limiter = AsyncRateLimiter(
            max_calls=59, period=60, callback=_rate_limit_handler
        )
        self.rate_limiters = AsyncRateLimiterManager(
            [self.global_rate_limiter, self.bot_rate_limiter]
        )
        self.user_agent = (
            f"topggpy (https://github.com/top-gg/python-sdk {__version__}) Python/"
            f"{sys.version_info[0]}.{sys.version_info[1]} aiohttp/{aiohttp.__version__}"
        )

    async def request(self, method: str, endpoint: str, **kwargs: Any) -> dict:
        """Handles requests to the API."""
        rate_limiters = (
            self.rate_limiters
            if endpoint.startswith("/bots")
            else self.global_rate_limiter
        )
        url = f"{self.BASE}{endpoint}"

        if not self.token:
            raise errors.UnauthorizedDetected("Top.gg API token not provided")

        headers = {
            "User-Agent": self.user_agent,
            "Content-Type": "application/json",
            "Authorization": self.token,
        }

        if "json" in kwargs:
            kwargs["data"] = to_json(kwargs.pop("json"))

        kwargs["headers"] = headers

        for _ in range(2):
            async with rate_limiters:
                async with self.session.request(method, url, **kwargs) as resp:
                    log.debug(
                        "%s %s with %s has returned %s",
                        method,
                        url,
                        kwargs.get("data"),
                        resp.status,
                    )

                    data = await _json_or_text(resp)

                    if 300 > resp.status >= 200:
                        return cast(dict, data)

                    elif resp.status == 429:  # we are being ratelimited
                        fmt = "We are being ratelimited. Retrying in %.2f seconds (%.3f minutes)."

                        # sleep a bit
                        retry_after = float(resp.headers.get("Retry-After"))
                        mins = retry_after / 60
                        log.warning(fmt, retry_after, mins)

                        # check if it's a global ratelimit (True as only 1 ratelimit atm - /api/bots)
                        # is_global = True
                        # is_global = data.get('global', False)
                        # if is_global:
                        #     self._global_over.clear()

                        await asyncio.sleep(retry_after, loop=self.loop)
                        log.debug("Done sleeping for the ratelimit. Retrying...")

                        # release the global lock now that the
                        # global ratelimit has passed
                        # if is_global:
                        #     self._global_over.set()
                        log.debug("Global ratelimit is now over.")
                        continue

                    elif resp.status == 400:
                        raise errors.HTTPException(resp, data)
                    elif resp.status == 401:
                        raise errors.Unauthorized(resp, data)
                    elif resp.status == 403:
                        raise errors.Forbidden(resp, data)
                    elif resp.status == 404:
                        raise errors.NotFound(resp, data)
                    elif resp.status >= 500:
                        raise errors.ServerError(resp, data)

        # We've run out of retries, raise.
        raise errors.HTTPException(resp, data)

    async def close(self) -> None:
        await self.session.close()

    async def post_guild_count(
        self,
        guild_count: Optional[Union[int, List[int]]],
        shard_count: Optional[int],
        shard_id: Optional[int],
    ) -> None:
        """Posts bot's guild count and shards info on Top.gg."""
        payload = {"server_count": guild_count}
        if shard_count:
            payload["shard_count"] = shard_count
        if shard_id:
            payload["shard_id"] = shard_id

        await self.request("POST", "/bots/stats", json=payload)

    def get_weekend_status(self) -> Coroutine[Any, Any, dict]:
        """Gets the weekend status from Top.gg."""
        return self.request("GET", "/weekend")

    def get_guild_count(self, bot_id: int) -> Coroutine[Any, Any, dict]:
        """Gets the guild count of the given Bot ID."""
        return self.request("GET", f"/bots/{bot_id}/stats")

    def get_bot_info(self, bot_id: int) -> Coroutine[Any, Any, dict]:
        """Gets the information of a bot under given bot ID on Top.gg."""
        return self.request("GET", f"/bots/{bot_id}")

    def get_bot_votes(self, bot_id: int) -> Coroutine[Any, Any, Iterable[dict]]:
        """Gets your bot's last 1000 votes on Top.gg."""
        return self.request("GET", f"/bots/{bot_id}/votes")

    def get_bots(
        self,
        limit: int,
        offset: int,
        sort: str,
        search: Dict[str, str],
        fields: Sequence[str],
    ) -> Coroutine[Any, Any, dict]:
        """Gets an object of bots on Top.gg."""
        limit = min(limit, 500)
        fields = ", ".join(fields)
        search = " ".join([f"{field}: {value}" for field, value in search.items()])

        return self.request(
            "GET",
            "/bots",
            params={
                "limit": limit,
                "offset": offset,
                "sort": sort,
                "search": search,
                "fields": fields,
            },
        )

    def get_user_info(self, user_id: int) -> Coroutine[Any, Any, dict]:
        """Gets an object of the user on Top.gg."""
        return self.request("GET", f"/users/{user_id}")

    def get_user_vote(self, bot_id: int, user_id: int) -> Coroutine[Any, Any, dict]:
        """Gets info whether the user has voted for your bot."""
        return self.request("GET", f"/bots/{bot_id}/check", params={"userId": user_id})


async def _rate_limit_handler(until: float) -> None:
    """Handles the displayed message when we are ratelimited."""
    duration = round(until - datetime.utcnow().timestamp())
    mins = duration / 60
    fmt = (
        "We have exhausted a ratelimit quota. Retrying in %.2f seconds (%.3f minutes)."
    )
    log.warning(fmt, duration, mins)


def to_json(obj: Any) -> str:
    if json.__name__ == "ujson":
        return json.dumps(obj, ensure_ascii=True)
    return json.dumps(obj, separators=(",", ":"), ensure_ascii=True)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2021 Assanali Mukhanov

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from aiohttp import ClientResponse


class TopGGException(Exception):
    """Base exception class for topggpy.

    Ideally speaking, this could be caught to handle any exceptions thrown from this library.
    """

    pass


class ClientException(TopGGException):
    """Exception that's thrown when an operation in the :class:`DBLClient` fails.

    These are usually for exceptions that happened due to user input.
    """

    pass


class HTTPException(TopGGException):
    """Exception that's thrown when an HTTP request operation fails.

    Attributes
    ----------
    response: :class:`aiohttp.ClientResponse`
        The response of the failed HTTP request.
    text: str
        The text of the error. Could be an empty string.
    """

    def __init__(self, response: "ClientResponse", message: Union[dict, str]) -> None:
        self.response = response
        if isinstance(message, dict):
            self.text = message.get("message", "")
            self.code = message.get("code", 0)
        else:
            self.text = message

        fmt = f"{self.response.reason} (status code: {self.response.status})"
        if self.text:
            fmt = f"{fmt}: {self.text}"

        super().__init__(fmt)


class Unauthorized(HTTPException):
    """Exception that's thrown when status code 401 occurs."""

    pass


class UnauthorizedDetected(TopGGException):
    """Exception that's thrown when no API Token is provided."""

    pass


class Forbidden(HTTPException):
    """Exception that's thrown when status code 403 occurs."""

    pass


class NotFound(HTTPException):
    """Exception that's thrown when status code 404 occurs."""

    pass


class ServerError(HTTPException):
    """Exception that's thrown when Top.gg returns "Server Error" responses (status codes such as 500 and 503)."""

    pass



# -*- coding: utf-8 -*-

"""
Top.gg Python API Wrapper
~~~~~~~~~~~~~~~~~~~~~~~~~
A basic wrapper for the Top.gg API.
:copyright: (c) 2021 Assanali Mukhanov & Top.gg
:license: MIT, see LICENSE for more details.
"""

from collections import namedtuple

__title__ = "topggpy"
__author__ = "Assanali Mukhanov"
__license__ = "MIT"
__version__ = "1.4.0"

VersionInfo = namedtuple("VersionInfo", "major minor micro releaselevel serial")
major, minor, micro = (int(i) for i in __version__.split("."))
version_info = VersionInfo(
    major=major, minor=minor, micro=micro, releaselevel="final", serial=0
)

from .client import DBLClient
from .errors import *
from .http import HTTPClient
from .types import WidgetOptions
from .webhook import WebhookManager



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2021 Assanali Mukhanov

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
import collections
from datetime import datetime
from types import TracebackType
from typing import Any, Awaitable, Callable, List, Optional, Type


class AsyncRateLimiter:
    """
    Provides rate limiting for an operation with a configurable number of requests for a time period.
    """

    __lock: asyncio.Lock
    callback: Optional[Callable[[float], Awaitable[Any]]]
    max_calls: int
    period: float
    calls: collections.deque

    def __init__(
        self,
        max_calls: int,
        period: float = 1.0,
        callback: Optional[Callable[[float], Awaitable[Any]]] = None,
    ):
        if period <= 0:
            raise ValueError("Rate limiting period should be > 0")
        if max_calls <= 0:
            raise ValueError("Rate limiting number of calls should be > 0")
        self.calls = collections.deque()

        self.period = period
        self.max_calls = max_calls
        self.callback = callback
        self.__lock = asyncio.Lock()

    async def __aenter__(self) -> "AsyncRateLimiter":
        async with self.__lock:
            if len(self.calls) >= self.max_calls:
                until = datetime.utcnow().timestamp() + self.period - self._timespan
                if self.callback:
                    asyncio.ensure_future(self.callback(until))
                sleep_time = until - datetime.utcnow().timestamp()
                if sleep_time > 0:
                    await asyncio.sleep(sleep_time)
            return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        async with self.__lock:
            # Store the last operation timestamp.
            self.calls.append(datetime.utcnow().timestamp())

            while self._timespan >= self.period:
                self.calls.popleft()

    @property
    def _timespan(self) -> float:
        return self.calls[-1] - self.calls[0]


class AsyncRateLimiterManager:
    rate_limiters: List[AsyncRateLimiter]

    def __init__(self, rate_limiters: List[AsyncRateLimiter]):
        self.rate_limiters = rate_limiters

    async def __aenter__(self) -> "AsyncRateLimiterManager":
        [await manager.__aenter__() for manager in self.rate_limiters]
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        [
            await manager.__aexit__(exc_type, exc_val, exc_tb)
            for manager in self.rate_limiters
        ]



import os
import pathlib
import re
import types

from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent

txt = (HERE / "topgg" / "__init__.py").read_text("utf-8")

groups = {}

for match in re.finditer(r'__(?P<identifier>.*)__\s*=\s*"(?P<value>[^"]+)"\r?', txt):
    group = match.groupdict()
    groups[group["identifier"]] = group["value"]

metadata = types.SimpleNamespace(**groups)

on_rtd = os.getenv("READTHEDOCS") == "True"

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

if on_rtd:
    requirements.append("sphinxcontrib-napoleon")
    requirements.append("sphinx-rtd-dark-mode")

with open("README.rst") as f:
    readme = f.read()

setup(
    name="topggpy",
    author=f"{metadata.author}, Top.gg",
    author_email="shivaco.osu@gmail.com",
    url="https://github.com/top-gg/python-sdk",
    version=metadata.version,
    packages=find_packages(),
    license=metadata.license,
    description="A simple API wrapper for Top.gg written in Python.",
    long_description=readme,
    package_data={"topgg": ["py.typed"]},
    include_package_data=True,
    python_requires=">= 3.6",
    install_requires=requirements,
    keywords="discord bot server list discordservers serverlist discordbots botlist topgg top.gg",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2021 Assanali Mukhanov

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
import logging
import sys
import traceback
from asyncio.tasks import Task
from contextlib import suppress
from typing import Any, Dict, List, Optional, Union, cast

import discord
from discord.ext.commands.bot import BotBase

# if sys.version_info >= (3, 8):
#     from typing import TypedDict
# else:
#     from typing_extensions import TypedDict
from . import errors, types
from .http import HTTPClient

log = logging.getLogger(__name__)


class DBLClient:
    """Represents a client connection that connects to Top.gg.
    This class is used to interact with the Top.gg API.

    .. _event loop: https://docs.python.org/3/library/asyncio-eventloops.html
    .. _aiohttp session: https://aiohttp.readthedocs.io/en/stable/client_reference.html#client-session

    Parameters
    ----------
    bot: discord.Client
        An instance of a discord.py Client object.
    token: str
        Your bot's Top.gg API Token.
    autopost: bool
        Whether to automatically post bot's guild count every 30 minutes.
        This will dispatch :meth:`on_autopost_success` (or :meth:`on_autopost_error` in case of an error).
    post_shard_count: bool
        Whether to post the shard count on autopost.
        Defaults to False.
    autopost_interval: Optional[int]
        Interval used by autopost to post server count automatically, measured in seconds. Defaults to 1800 (30
        minutes) if autopost is True, otherwise None.
    **session: :class:`aiohttp.ClientSession`
        An `aiohttp session`_ to use for requests to the API.
    """

    bot: discord.Client
    bot_id: int
    loop: asyncio.AbstractEventLoop
    autopost: bool
    post_shard_count: bool
    _is_closed: bool
    http: HTTPClient
    autopost_task: Task
    autopost_interval: Optional[float]

    def __init__(
        self,
        bot: discord.Client,
        token: str,
        autopost: bool = False,
        post_shard_count: bool = False,
        autopost_interval: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        self.bot = bot
        self.loop = bot.loop
        self.autopost = autopost
        self.post_shard_count = post_shard_count
        self.autopost_interval = autopost_interval
        self.http = HTTPClient(token, loop=self.loop, session=kwargs.get("session"))
        self._is_closed = False

        if not isinstance(self.autopost, bool):
            raise errors.ClientException("autopost must be of type bool")
        if not isinstance(self.post_shard_count, bool):
            raise errors.ClientException("post_shard_count must be of type bool")
        if self.autopost:
            if self.autopost_interval is None:
                self.autopost_interval = 1800
            if not isinstance(self.autopost_interval, int):
                raise errors.ClientException("autopost_interval must be of type int")
            if self.autopost_interval < 900:
                raise errors.ClientException(
                    "autopost_interval must be greater than or equal to 900 seconds (15 minutes)"
                )

            if not hasattr(self.bot, "on_autopost_error"):
                self.bot.on_autopost_error = self.on_autopost_error

            self.autopost_task = self.loop.create_task(self._auto_post())
        else:
            if self.post_shard_count:
                raise errors.ClientException(
                    "autopost must be activated if post_shard_count is passed"
                )
            if self.autopost_interval:
                raise errors.ClientException(
                    "autopost must be activated if autopost_interval is passed"
                )

    async def on_autopost_error(self, exception: Exception) -> None:
        # only print if there's no external autopost_error listeners.
        if isinstance(self.bot, BotBase) and self.bot.extra_events.get(
            "on_autopost_error"
        ):
            return

        print("Ignoring exception in auto post loop:", file=sys.stderr)
        traceback.print_exception(
            type(exception), exception, exception.__traceback__, file=sys.stderr
        )

    async def _ensure_bot_user(self) -> None:
        await self.bot.wait_until_ready()
        if not hasattr(self, "bot_id") or self.bot_id is None:
            self.bot_id = self.bot.user.id

    async def _auto_post(self) -> None:
        await self._ensure_bot_user()
        while not self.bot.is_closed():
            try:
                log.debug(f"Attempting to post server count ({self.guild_count})")
                await self.post_guild_count(
                    shard_count=self.bot.shard_count if self.post_shard_count else None
                )
                event_name = "autopost_success"
                self.bot.dispatch(event_name)
            except Exception as e:
                event_name = "autopost_error"
                self.bot.dispatch(event_name, e)

                if isinstance(e, errors.Unauthorized):
                    raise

            await asyncio.sleep(cast(float, self.autopost_interval))

    @property
    def is_closed(self) -> bool:
        return self._is_closed

    @property
    def guild_count(self) -> int:
        """Gets the guild count from the provided Client object."""
        return len(self.bot.guilds)

    async def get_weekend_status(self) -> bool:
        """This function is a coroutine.

        Gets weekend status from Top.gg.

        Returns
        -------
        weekend status: bool
            The boolean value of weekend status.
        """
        data = await self.http.get_weekend_status()
        return data["is_weekend"]

    async def post_guild_count(
        self,
        guild_count: Optional[Union[int, List[int]]] = None,
        shard_count: Optional[int] = None,
        shard_id: Optional[int] = None,
    ) -> None:
        """This function is a coroutine.

        Posts your bot's guild count and shards info to Top.gg.

        .. _0 based indexing : https://en.wikipedia.org/wiki/Zero-based_numbering

        Parameters
        ----------
        guild_count: Optional[Union[int, List[int]]]
            Number of guilds the bot is in. Applies the number to a shard instead if shards are specified.
            If not specified, length of provided client's property `.guilds` will be posted.
        shard_count: Optional[int]
            The total number of shards.
        shard_id: Optional[int]
            The index of the current shard. Top.gg uses `0 based indexing`_ for shards.
        """
        await self._ensure_bot_user()
        if guild_count is None:
            guild_count = self.guild_count
        await self.http.post_guild_count(guild_count, shard_count, shard_id)

    async def get_guild_count(self, bot_id: Optional[int] = None) -> types.BotStatsData:
        """This function is a coroutine.

        Gets a bot's guild count and shard info from Top.gg.

        Parameters
        ----------
        bot_id: int
            ID of the bot you want to look up. Defaults to the provided Client object.

        Returns
        -------
        stats: :ref:`BotStatsData`
            The guild count and shards of a bot on Top.gg.
        """
        await self._ensure_bot_user()
        if bot_id is None:
            bot_id = self.bot_id
        response = await self.http.get_guild_count(bot_id)
        return types.BotStatsData(**response)

    async def get_bot_votes(self) -> List[types.BriefUserData]:
        """This function is a coroutine.

        Gets information about last 1000 votes for your bot on Top.gg.

        .. note::
            This API endpoint is only available to the bot's owner.

        Returns
        -------
        users: List[:ref:`BriefUserData`]
            Users who voted for your bot.
        """
        await self._ensure_bot_user()
        response = await self.http.get_bot_votes(self.bot_id)
        return [types.BriefUserData(**user) for user in response]

    async def get_bot_info(self, bot_id: Optional[int] = None) -> types.BotData:
        """This function is a coroutine.

        Gets information about a bot from Top.gg.

        Parameters
        ----------
        bot_id: int
            ID of the bot to look up. Defaults to the provided Client object.

        Returns
        -------
        bot info: :ref:`BotData`
            Information on the bot you looked up. Returned data can be found
            `here <https://docs.top.gg/api/bot/#bot-structure>`_.
        """
        await self._ensure_bot_user()
        if bot_id is None:
            bot_id = self.bot_id
        response = await self.http.get_bot_info(bot_id)
        return types.BotData(**response)

    async def get_bots(
        self,
        limit: int = 50,
        offset: int = 0,
        sort: Optional[str] = None,
        search: Optional[Dict[str, Any]] = None,
        fields: Optional[List[str]] = None,
    ) -> types.DataDict[str, Any]:
        """This function is a coroutine.

        Gets information about listed bots on Top.gg.

        Parameters
        ----------
        limit: int
            The number of results to look up. Defaults to 50. Max 500 allowed.
        offset: int
            The amount of bots to skip. Defaults to 0.
        sort: str
            The field to sort by. Prefix with ``-`` to reverse the order.
        search: Dict[str, Any]
            The search data.
        fields: List[str]
            Fields to output.

        Returns
        -------
        bots: :ref:`DataDict`
            Info on bots that match the search query on Top.gg.
        """
        sort = sort or ""
        search = search or {}
        fields = fields or []
        response = await self.http.get_bots(limit, offset, sort, search, fields)
        response["results"] = [
            types.BotData(**bot_data) for bot_data in response["results"]
        ]
        return types.DataDict(**response)

    async def get_user_info(self, user_id: int) -> types.UserData:
        """This function is a coroutine.

        Gets information about a user on Top.gg.

        Parameters
        ----------
        user_id: int
            ID of the user to look up.

        Returns
        -------
        user data: :ref:`UserData`
            Information about a Top.gg user.
        """
        response = await self.http.get_user_info(user_id)
        return types.UserData(**response)

    async def get_user_vote(self, user_id: int) -> bool:
        """This function is a coroutine.

        Gets information about a user's vote for your bot on Top.gg.

        Parameters
        ----------
        user_id: int
            ID of the user.

        Returns
        -------
        vote status: bool
            Info about the user's vote.
        """
        await self._ensure_bot_user()
        data = await self.http.get_user_vote(self.bot_id, user_id)
        return bool(data["voted"])

    async def generate_widget(self, options: types.WidgetOptions) -> str:
        """This function is a coroutine.

        Generates a Top.gg widget from the provided :ref:`WidgetOptions` object.

        Parameters
        ----------
        options: :ref:`WidgetOptions`
            A :ref:`WidgetOptions` object containing widget parameters.

        Returns
        -------
        widget: str
            Generated widget URL.
        """
        if not isinstance(options, types.WidgetOptions):
            raise errors.ClientException(
                "options argument passed to generate_widget must be of type WidgetOptions"
            )
        bot_id = options.id

        if bot_id is None:
            await self._ensure_bot_user()
            bot_id = self.bot_id

        widget_query = f"noavatar={str(options.noavatar).lower()}"
        for key, value in options.colors.items():
            widget_query += f"&{key.lower()}{'' if key.lower().endswith('color') else 'color'}={value:x}"
        widget_format = options.format
        widget_type = f"/{options.type}" if options.type else ""

        url = f"""https://top.gg/api/widget{widget_type}/{bot_id}.{widget_format}?{widget_query}"""
        return url

    async def close(self) -> None:
        """This function is a coroutine.

        Closes all connections.
        """
        if self.is_closed:
            return
        else:
            await self.http.close()
            if self.autopost:
                self.autopost_task.cancel()
            self._is_closed = True

        with suppress(AttributeError):
            delattr(self.bot, "on_autopost_error")
